#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/stat.h>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

struct device_info {
	cl_device_id id;
	cl_device_type type;
	unsigned int units;
	unsigned int clock;

	unsigned int dim;
	size_t *work_item_sizes;
	size_t work_group_size;

	unsigned long mem_size;
};

static char *load_source(const char *fname);
static int select_device(struct device_info *di, int (*devcmp)(struct device_info*, struct device_info*));
static int get_dev_info(cl_device_id dev, struct device_info *di);
static int devcmp(struct device_info *a, struct device_info *b);
static const char *devtypestr(cl_device_type type);
static void print_memsize(FILE *out, unsigned long memsz);

#define MIN(a, b)	((a) < (b) ? (a) : (b))

int main(void)
{
	struct device_info devinf;
	cl_context ctx;
	cl_command_queue cmdq;
	cl_mem inbuf, outbuf;
	cl_program prog;
	cl_kernel ker;
	size_t global_size[] = {1, 1, 1}, local_size[] = {1, 1, 1};
	int i, err;
	int data[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};
	int count = sizeof data / sizeof *data;
	char *src_buf;

	if(select_device(&devinf, devcmp) == -1) {
		return 1;
	}

	ctx = clCreateContext(0, 1, &devinf.id, 0, 0, &err);
	if(err) {
		fprintf(stderr, "error creating cl context\n");
		return 1;
	}
	cmdq = clCreateCommandQueue(ctx, devinf.id, 0, &err);
	if(err) {
		fprintf(stderr, "error creating cl command queue\n");
		return 1;
	}

	inbuf = clCreateBuffer(ctx, CL_MEM_READ_ONLY, sizeof data, 0, &err);
	if(err) {
		fprintf(stderr, "error creating buffer\n");
		return 1;
	}
	outbuf = clCreateBuffer(ctx, CL_MEM_WRITE_ONLY, sizeof data, 0, &err);
	if(err) {
		fprintf(stderr, "error creating buffer\n");
		return 1;
	}


	for(i=0; i<count; i++) {
		printf("%3d ", data[i]);
	}
	putchar('\n');

	if(clEnqueueWriteBuffer(cmdq, inbuf, 1, 0, sizeof data, data, 0, 0, 0) != 0) {
		fprintf(stderr, "error writing into input buffer\n");
		return 1;
	}


	if(!(src_buf = load_source("test.cl"))) {
		return 1;
	}

	prog = clCreateProgramWithSource(ctx, 1, (const char**)&src_buf, 0, &err);
	if(err) {
		fprintf(stderr, "error compiling kernel\n");
		return 1;
	}

	if(clBuildProgram(prog, 0, 0, 0, 0, 0) != 0) {
		fprintf(stderr, "failed to build program\n");
		return 1;
	}

	ker = clCreateKernel(prog, "test", &err);
	if(err) {
		fprintf(stderr, "failed to create kernel\n");
		return 1;
	}

	if(clSetKernelArg(ker, 0, sizeof inbuf, &inbuf) != 0) {
		fprintf(stderr, "clSetKernelArg 0 failed\n");
		return 1;
	}
	if(clSetKernelArg(ker, 1, sizeof outbuf, &outbuf) != 0) {
		fprintf(stderr, "clSetKernelArg 1 failed\n");
		return 1;
	}

	global_size[0] = count;
	local_size[0] = 1;

	if((err = clEnqueueNDRangeKernel(cmdq, ker, 1, 0, global_size, 0, 0, 0, 0)) != 0) {
		fprintf(stderr, "error executing kernel: %d\n", err);
		return 1;
	}


	if(clEnqueueReadBuffer(cmdq, outbuf, 1, 0, sizeof data, data, 0, 0, 0) != 0) {
		fprintf(stderr, "error reading output buffer\n");
		return 1;
	}

	for(i=0; i<count; i++) {
		printf("%3d ", data[i]);
	}
	putchar('\n');

	clReleaseMemObject(inbuf);
	clReleaseMemObject(outbuf);
	free(devinf.work_item_sizes);
	free(src_buf);
	return 0;
}


static char *load_source(const char *fname)
{
	FILE *fp;
	char *src;
	struct stat st;

	if(!(fp = fopen(fname, "rb"))) {
		fprintf(stderr, "failed to open file: %s: %s\n", fname, strerror(errno));
		return 0;
	}

	fstat(fileno(fp), &st);

	if(!(src = malloc(st.st_size + 1))) {
		fclose(fp);
		return 0;
	}

	fread(src, 1, st.st_size, fp);
	src[st.st_size] = 0;
	fclose(fp);

	return src;
}

static int select_device(struct device_info *dev_inf, int (*devcmp)(struct device_info*, struct device_info*))
{
	unsigned int i, j, num_dev, sel;
	cl_device_id dev[32];

	dev_inf->work_item_sizes = 0;

	clGetDeviceIDs(0, CL_DEVICE_TYPE_ALL, 32, dev, &num_dev);
	printf("found %d cl devices.\n", num_dev);

	for(i=0; i<num_dev; i++) {
		struct device_info di;

		if(get_dev_info(dev[i], &di) == -1) {
			free(dev_inf->work_item_sizes);
			return -1;
		}

		printf("--> device %u (%s)\n", i, devtypestr(di.type));
		printf("max compute units: %u\n", di.units);
		printf("max clock frequency: %u\n", di.clock);
		printf("max work item dimensions: %u\n", di.dim);

		printf("max work item sizes: ");
		for(j=0; j<di.dim; j++) {
			printf("%u", (unsigned int)di.work_item_sizes[j]);
			if(di.dim - j > 1) {
				printf(", ");
			}
		}
		putchar('\n');

		printf("max work group size: %u\n", (unsigned int)di.work_group_size);
		printf("max object allocation size: ");
		print_memsize(stdout, di.mem_size);
		putchar('\n');

		if(devcmp(&di, dev_inf) > 0) {
			free(dev_inf->work_item_sizes);
			memcpy(dev_inf, &di, sizeof di);
			sel = i;
		}
	}

	if(num_dev) {
		printf("\nusing device: %d\n", sel);
		return 0;
	}

	return -1;
}

static int get_dev_info(cl_device_id dev, struct device_info *di)
{
	di->id = dev;

	clGetDeviceInfo(dev, CL_DEVICE_TYPE, sizeof di->type, &di->type, 0);
	clGetDeviceInfo(dev, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof di->units, &di->units, 0);
	clGetDeviceInfo(dev, CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof di->clock, &di->clock, 0);
	clGetDeviceInfo(dev, CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS, sizeof di->dim, &di->dim, 0);

	if(!(di->work_item_sizes = malloc(di->dim * sizeof *di->work_item_sizes))) {
		perror("malloc failed");
		return -1;
	}

	clGetDeviceInfo(dev, CL_DEVICE_MAX_WORK_ITEM_SIZES, di->dim * sizeof *di->work_item_sizes, di->work_item_sizes, 0);
	clGetDeviceInfo(dev, CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof di->work_group_size, &di->work_group_size, 0);
	clGetDeviceInfo(dev, CL_DEVICE_MAX_MEM_ALLOC_SIZE, sizeof di->mem_size, &di->mem_size, 0);

	return 0;
}

static int devcmp(struct device_info *a, struct device_info *b)
{
	unsigned int aval = a->units * a->clock;
	unsigned int bval = b->units * b->clock;

	return aval - bval;
}

static const char *devtypestr(cl_device_type type)
{
	switch(type) {
	case CL_DEVICE_TYPE_CPU:
		return "cpu";
	case CL_DEVICE_TYPE_GPU:
		return "gpu";
	case CL_DEVICE_TYPE_ACCELERATOR:
		return "accelerator";
	default:
		break;
	}
	return "unknown";
}

static void print_memsize(FILE *out, unsigned long bytes)
{
	int i;
	unsigned long memsz = bytes;
	char *suffix[] = {"bytes", "kb", "mb", "gb", "tb", "pb", 0};

	for(i=0; suffix[i]; i++) {
		if(memsz < 1024) {
			fprintf(out, "%lu %s", memsz, suffix[i]);
			if(i > 0) {
				fprintf(out, " (%lu bytes)", bytes);
			}
			return;
		}

		memsz /= 1024;
	}
}
